/*
 * com.gregor.rrd.RRDTool
 *
 * $Id: RRDTool.java,v 1.1.1.1 2004/08/17 01:07:04 djgregor Exp $
 *
 * Copyright (c) 2003 Daniel J. Gregor, Jr.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package com.gregor.rrd;

import java.io.IOException;
import java.io.OutputStream;

import java.text.CharacterIterator;
import java.text.StringCharacterIterator;

import java.util.Map;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;

import com.gregor.util.PropertyGetter;
import com.gregor.util.PropertySetter;
import com.gregor.util.PropertySetterException;

import java.beans.IntrospectionException;

/**
 * Interface to Round Robin Databases (RRD's).
 */
public abstract class RRDTool {
	/**
	 * Initializes the RRD interface.
	 *
	 * Performs implementation-specific initialization, such as starting
	 * a copy of rrdtool for an implementation over pipes, or calling a
	 * native initialization function for a JNI implementation.
	 */
	public abstract void init() throws IOException, RRDException;

	/**
	 * De-initializes the RRD interface.
	 *
	 * Explicitly destroys any resources in use, such as memory or external
	 * processes.
	 */
	public abstract void deInit() throws IOException, RRDException;

	/**
	 * Runs an RRD command with the specified command string.
	 *
	 * @param command command string, as would be passed to rrdtool on stdin
	 * @return return value from RRD command, one array element per line
	 * @throws IOException If an IO error occurs in the implementation driver
	 * @throws RRDException If an RRD-related problem occurs in the
	 * implementation driver
	 */
	public String[] doCommand(String command) throws IOException, RRDException {
		return doCommand(command, null);
	}

	/**
	 * Runs an RRD command with the specified pre-parsed command string array.
	 *
	 * @param command pre-parsed command string array, one argument per
	 * array element
	 * @return return value from RRD command, one array element per line
	 * @throws IOException If an IO error occurs in the implementation driver
	 * @throws RRDException If an RRD-related problem occurs in the
	 * implementation driver
	 */
	public String[] doCommand(String[] command) throws IOException,
			RRDException {
		return doCommand(command, null);
	}

	/**
	 * Runs an RRD command with the specified command string.
	 *
	 * @param command command string, as would be passed to rrdtool on stdin
	 * @param out stream for graph data output (if used), null otherwise
	 * @return return value from RRD command, one array element per line
	 * @throws IOException If an IO error occurs in the implementation driver
	 * @throws RRDException If an RRD-related problem occurs in the
	 * implementation driver
	 */
	public abstract String[] doCommand(String command, OutputStream out)
			throws IOException, RRDException;

	/**
	 * Runs an RRD command with the specified pre-parsed command string array.
	 *
	 * @param command pre-parsed command string array, one argument per
	 * array element
	 * @param out stream for graph data output (if used), null otherwise
	 * @return return value from RRD command, one array element per line
	 * @throws IOException If an IO error occurs in the implementation driver
	 * @throws RRDException If an RRD-related problem occurs in the
	 * implementation driver
	 */
	public abstract String[] doCommand(String[] command, OutputStream out)
			throws IOException, RRDException;

	/**
	 * Creates a new instance of an RRDTool implementation by class name.
	 *
	 * @param className Class name to load and instantiate
	 * @return Instantiated RRDTool implementation
	 * @throws ClassNotFoundException If the specific class is not found by
	 * Class.forName
	 * @throws InstantiationException If the class is not an instantiatable
	 * class by Class.newInstance
	 * @throws IllegalAccessException If the class or its nullary constructor
	 * is not accessable by Class.newInstance
	 * @see Class#forName
	 * @see Class#newInstance
	 */
	public static RRDTool newInstanceOf(String className)
			throws ClassNotFoundException, InstantiationException,
			IllegalAccessException {
		return (RRDTool) Class.forName(className).newInstance();
	}

	/**
	 * Sets a property value using the implementation's setter methods.
	 *
	 * @param name Name of property value
	 * @param value Value to set
	 * @throws RRDException If an error occurs setting the value
	 */
	public void setProperty(String name, String value) throws RRDException {
		try {
			PropertySetter ps = new PropertySetter(this);
			ps.setProperty(name, value);
		} catch (PropertySetterException e) {
			throw new RRDException(e.toString());
		}
	}

	/**
	 * Sets a group of property value using the implementation's setter
	 * methods.
	 *
	 * The properties string is a collection of key=value pairs, separated
	 * by semicolons (";");
	 *
	 * @param properties Properties string
	 * @throws RRDException If an error occurs setting the value or parsing
	 * the properties.
	 */

	public void setProperties(String properties) throws RRDException {
		StringTokenizer props = new StringTokenizer(properties, ";");

		while (props.hasMoreTokens()) {
			String token = props.nextToken();
			int x;

			if ((x = token.indexOf('=')) == -1) {
				throw new RRDException("could not parse property "
						+ "key=value pair");
			}

			setProperty(token.substring(0, x), token.substring(x + 1));
		}
	}

	/**
	 * Gets a property value using the implementation's getter methods.
	 *
	 * @param name Name of property value to retrieve
	 * @return Property value
	 * @throws RRDException If an error occurs getting the value
	 */
	public String getProperty(String name) throws RRDException {
		try {
			return PropertyGetter.getProperty(this, name);
		} catch (IntrospectionException e) {
			throw new RRDException(e.toString());
		}
	}

	/**
	 * Creates a new RRD file with the specified argument string.
	 *
	 * @param filename The RRD file to create
	 * @param create Arguments to rrdcreate.  Includes any options
	 * (--start, --step, etc.) and the datasource and round-robin
	 * archive definitions
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	public void create(String filename, String create) throws IOException,
			RRDException {
		doCommand("create " + escape(filename) + " " + create);
	}

	/**
	 * Creates a new RRD file with the specified parameters.
	 *
	 * The dataSource parameter is a Map whose keys contain the
	 * data source definitions (DS:...) and values contain arrays
	 * of round-robin archives (RRA:...) for each corresponding
	 * data source key.
	 *
	 * @param filename The RRD file to create
	 * @param start Start time in milliseconds since the epoch
	 * @param step Step between samples in milliseconds
	 * @param dataSources Mapping between data source and RRA definitions
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	public void create(String filename, long start, long step, Map dataSources)
			throws IOException, RRDException {

		LinkedList command = new LinkedList();

		command.add("create");
		command.add(filename);
		command.add("--start");
		command.add(Long.toString(start / 1000));
		command.add("--step");
		command.add(Long.toString(step / 1000));

		for (Iterator i = dataSources.keySet().iterator(); i.hasNext();) {
			String dataSource = (String) i.next();
			String[] archives = (String[]) dataSources.get(dataSource);

			command.add(dataSource);
			for (int a = 0; a < archives.length; a++) {
				command.add(archives[a]);
			}
		}

		doCommand((String[]) command.toArray(new String[0]));
	}

	/**
	 * Fetches data from an RRD file with default resolution and time values.
	 *
	 * @param filename Fetch data from this RRD file
	 * @param CF Consolidation function
	 * @return Results
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	public FetchResults fetch(String filename, String CF) throws IOException,
			RRDException {
		String[] command = new String[3];
		command[0] = "fetch";
		command[1] = filename;
		command[2] = CF;

		return doFetch(command);
	}

	/**
	 * Fetches data from an RRD file with specified resolution and
	 * default time values.
	 *
	 * @param filename Fetch data from this RRD file
	 * @param CF Consolidation function
	 * @param resolution Attempt to fetch data at this resolution, in
	 * milliseconds
	 * @return Results
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	public FetchResults fetch(String filename, String CF, long resolution)
			throws IOException, RRDException {
		String[] command = new String[5];
		command[0] = "fetch";
		command[1] = filename;
		command[2] = CF;
		command[3] = "resolution";
		command[4] = Long.toString(resolution / 1000);

		return doFetch(command);
	}

	/**
	 * Fetches data from an RRD file with default resolution and
	 * specified time values.
	 *
	 * @param filename Fetch data from this RRD file
	 * @param CF Consolidation function
	 * @param start Start time in milliseconds since the epoch
	 * @param end End time in milliseconds since the epoch
	 * @return Results
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	public FetchResults fetch(String filename, String CF, long start, long end)
			throws IOException, RRDException {
		String[] command = new String[7];
		command[0] = "fetch";
		command[1] = filename;
		command[2] = CF;
		command[3] = "--start";
		command[4] = Long.toString(start / 1000);
		command[5] = "--end";
		command[6] = Long.toString(end / 1000);

		return doFetch(command);
	}

	/**
	 * Fetches data from an RRD file with specified resolution and time
	 * values.
	 *
	 * @param filename Fetch data from this RRD file
	 * @param CF Consolidation function
	 * @param resolution Attempt to fetch data at this resolution, in
	 * milliseconds
	 * @param start Start time in milliseconds since the epoch
	 * @param end End time in milliseconds since the epoch
	 * @return Results
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	public FetchResults fetch(String filename, String CF, long resolution,
			long start, long end) throws IOException, RRDException {
		String[] command = new String[9];
		command[0] = "fetch";
		command[1] = filename;
		command[2] = CF;
		command[7] = "--resolution";
		command[8] = Long.toString(resolution / 1000);
		command[3] = "--start";
		command[4] = Long.toString(start / 1000);
		command[5] = "--end";
		command[6] = Long.toString(end / 1000);

		return doFetch(command);
	}

	/**
	 * Fetches and parses results using the specified command.
	 *
	 * @param command Array of command arguments
	 * @return Results
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	private FetchResults doFetch(String[] command) throws IOException,
			RRDException {
		String[] ret = doCommand(command);

		FetchResults r = new FetchResults();
		for (int i = 0; i < ret.length; i++) {
			r.parse(ret[i]);
		}

		return r;
	}

	/**
	 * Creates a graph using the specified graph command.
	 *
	 * The first element in the returned contains the graph size in pixels,
	 * and can be parsed by parseGraphSize.  The remaining elements are
	 * the output of PRINT commands.
	 *
	 * @param file Output file
	 * @param command Arguments to rrdgraph
	 * @return Graph results array
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 * @see #parseGraphSize
	 */
	public String[] graph(String file, String command) throws IOException,
			RRDException {
		String[] graph_file = { "graph", file };
		String[] ret = doCommand(escapeAndJoinCommand(graph_file) + command);
		if (ret.length < 1) {
			throw new RRDException("doCommand(\"graph ...\") returned too "
					+ "few output lines");
		}

		return ret;
	}

	public String[] graph(OutputStream out, String command) throws IOException,
			RRDException {
		String[] ret = doCommand("graph " + command, out);
		if (ret.length < 1) {
			throw new RRDException("doCommand(\"graph ...\") returned too "
					+ "few output lines");
		}

		return ret;
	}

	/**
	 * Parses graph size as returned by graph.
	 *
	 * The first returned element is the X size and the second is the
	 * Y size, in pixels.
	 *
	 * @param size The graph size as returned by the first array element
	 * from the graph command
	 * @return Graph size
	 * @throws RRDException If the size string could not be parsed
	 * @see #graph
	 */
	public int[] parseGraphSize(String size) throws RRDException {
		int[] sizes = new int[2];
		int x;

		if ((x = size.indexOf('x')) == -1) {
			throw new RRDException("could not parse image size from \"" + size
					+ "\"");
		}

		try {
			sizes[0] = Integer.parseInt(size.substring(0, x));
			sizes[1] = Integer.parseInt(size.substring(x + 1));
		} catch (NumberFormatException e) {
			throw new RRDException("could not parse image size due to "
					+ "NumberFormatException: " + e);
		}

		return sizes;
	}

	/**
	 * Gets the last update time of the RRD file.
	 *
	 * @param filename RRD file
	 * @return Last update time in milliseconds since the epoch
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 */
	public long last(String filename) throws IOException, RRDException {
		String[] command = new String[2];
		command[0] = "last";
		command[1] = filename;

		String ret[] = doCommand(command);

		if (ret.length != 1) {
			throw new RRDException("last returned an incorrect number of "
					+ "results");
		}

		try {
			return Long.parseLong(ret[0]) * 1000;
		} catch (NumberFormatException e) {
			throw new RRDException("last returned a non-number, \"" + ret[0]
					+ "\": " + e);
		}
	}

	/**
	 * Updates an RRD with new data.
	 *
	 * This only updates for one timestamp at a time.  Each Map key is
	 * the name of a data source to update, and the corresponding value
	 * is the numerical value.
	 *
	 * @param filename RRD file
	 * @param timestamp data acquisition time
	 * @param values data
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 * @see Double
	 */
	public void update(String filename, long timestamp, Map values)
			throws IOException, RRDException {
		StringBuffer template = new StringBuffer();
		StringBuffer data = new StringBuffer();

		if (values.size() == 0) {
			throw new RRDException("No values in map to update");
		}

		for (Iterator i = values.keySet().iterator(); i.hasNext();) {
			String ds = (String) i.next();

			template.append(":");
			template.append(ds);

			data.append(":");
			data.append(values.get(ds).toString());
		}

		template.deleteCharAt(0);
		data.deleteCharAt(0);

		update(filename, timestamp, template.toString(), data.toString());
	}

	/**
	 * Updates an RRD with new data.
	 *
	 * This only updates for one timestamp at a time.  One data source
	 * and value can be specified or a list of each can be specified,
	 * each separated by colons.
	 *
	 * @param filename RRD file
	 * @param timestamp data acquisition time
	 * @param dataSources a single data source or a list of data sources
	 *                    separated by colons
	 * @param values a single value or a list of values separated by colons
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 * @see Double
	 */
	public void update(String filename, long timestamp, String dataSources,
			String values) throws IOException, RRDException {
		String[] command = new String[5];
		command[0] = "update";
		command[1] = filename;
		command[2] = "--template";
		command[3] = dataSources;
		command[4] = Long.toString(timestamp / 1000) + ":" + values;

		doCommand(command);
	}

	/**
	 * Updates an RRD with new data.
	 *
	 * This only updates for one timestamp at a time.  One data source
	 * and one value is specified.
	 *
	 * @param filename RRD file
	 * @param timestamp data acquisition time
	 * @param dataSource a single data source
	 * @param value a single value
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 * @see Double
	 */
	public void update(String filename, long timestamp, String dataSource,
			double value) throws IOException, RRDException {
		update(filename, timestamp, dataSource, Double.toString(value));
	}

	/**
	 * Updates an RRD with new data.
	 *
	 * This only updates for one timestamp at a time.  One data source
	 * and one value is specified.
	 *
	 * @param filename RRD file
	 * @param timestamp data acquisition time
	 * @param dataSource a single data source
	 * @param value a single value
	 * @throws IOException If an IO error occurs when calling doCommand
	 * @throws RRDException If an RRDException occurs when calling doCommand
	 * @see #doCommand
	 * @see Double
	 */
	public void update(String filename, long timestamp, String dataSource,
			long value) throws IOException, RRDException {
		update(filename, timestamp, dataSource, Long.toString(value));
	}

	/**
	 * Escapes and joins a command string array into a single string.
	 *
	 * @param command command string array
	 * @return joined command string
	 */
	public static String escapeAndJoinCommand(String[] command) {
		StringBuffer s = new StringBuffer();

		for (int i = 0; i < command.length; i++) {
			s.append(escape(command[i]));
			s.append(" ");
		}

		return s.toString();
	}

	/**
	 * Escapes characters in a command string.
	 *
	 * Escapes whitespace and quote characters in a command string and
	 * quotes the newly escaped string if needed.
	 *
	 * @param command command string
	 * @return escaped command string
	 */
	public static String escape(String command) {
		StringCharacterIterator iter = new StringCharacterIterator(command);

		StringBuffer a = new StringBuffer();

		boolean hasSpace = false;

		for (char c = iter.first(); c != CharacterIterator.DONE;
				c = iter	.next()) {

			switch (c) {
			case ' ':
				hasSpace = true;
				a.append(' ');
				break;

			case '\t':
				hasSpace = true;
				a.append("\\t");
				break;

			case '\n':
				hasSpace = true;
				a.append("\\n");
				break;

			case '\r':
				hasSpace = true;
				a.append("\\r");
				break;

			case '\\':
				a.append("\\\\");
				break;

			case '"':
				a.append("\\\"");
				break;

			case '\'':
				a.append("\\'");
				break;

			default:
				a.append(c);
				break;
			}

		}

		if (hasSpace) {
			return ("\"" + a.toString() + "\"");
		} else {
			return (a.toString());
		}
	}

	/**
	 * Deallocates resources.
	 *
	 * Simply calls deInit and ignores exceptions.
	 *
	 * @see #deInit
	 */
	protected void finalize() {
		try {
			deInit();
		} catch (IOException e) {
			// do nothing
		} catch (RRDException e) {
			// do nothing
		}
	}
	
	public static void main(String[] args) {
		try {
			String className = "com.gregor.rrd.RRDJRobin";
		
			RRDTool instance = RRDTool.newInstanceOf(className);
			String[] vals = instance.doCommand(args, System.out);
		
			if (vals != null) {
				for (int i = 0; i < vals.length; i++) {
					System.out.println(vals[i]);
				}
			}
		
			instance.deInit();
			System.exit(0);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			System.exit(1);
		}
	}
}